src_0_en = [
    "Have you noticed how accurate",
    "LLM are now that GPU have became more powerful, ",
    "especially when we think at the difficulties we had in the 2010 era",
    "where online chatbots were not performant",
    "at all, and ofter doing strict rules worked better",
    "do you remember that era?"]

src_0_fr = [
    "As-tu remarqué à quel point",
    "les LLM sont précis maintenant que les GPU sont devenus plus puissants, ",
    "surtout quand on pense aux difficultés qu'on avait dans les années 2010",
    "où les chatbots en ligne n'étaient pas performants",
    "du tout, et souvent faire des règles strictes fonctionnait mieux.",
    "Tu te souviens de cette époque, ",
    "c'était bien plus compliqué de travailler"]

src_1_fr = ["Il s’arrêtait par moments devant une villa",
    "coquettement nichée dans la verdure, il regardait",
    "par la grille et voyait au loin des femmes",
    "élégantes sur les balcons et les terrasses, des",
    "enfants couraient dans les jardins. Il s’intéressait",
    "surtout aux fleurs ; c’étaient elles qui attiraient",
    "particulièrement ses regards. De temps en temps,"
    "il voyait passer des cavaliers, des amazones et de"
    "belles voitures ; il les suivait d’un œil curieux et",
    "les oubliait avant qu’ils eussent disparu. "]

src_1_en = ["He would stop occasionally in front of a villa",
    "nestled charmingly in the greenery, look",
    "through the gate, and see elegant women",
    "on balconies and terraces in the distance,",
    "children running in the gardens. He was particularly interested",
    "in the flowers; they were what caught his eye. From time to time,", 
    "he would see horsemen, horsewomen, and beautiful carriages passing by;",
    "he would follow them with a curious eye and",
    "forget them before they had disappeared.",
    ]

src_2_fr = [
    "Il s'arrêtait par moments devant une villa",
    "coquettement nichée dans la verdure, il regardait",
    "par la grille et voyait au loin des femmes",
    "élégantes sur les balcons et les terrasses, des",
    "enfants couraient dans les jardins. Il s'intéressait",
    "surtout aux fleurs ; c'étaient elles qui attiraient",
    "particulièrement ses regards. De temps en temps,",
    "il voyait passer des cavaliers, des amazones et de",
    "belles voitures ; il les suivait d'un œil curieux et",
    "les oubliait avant qu'ils eussent disparu. ",
    "Une fois, il s'arrêta et compta son argent ; il",
    "lui restait trente kopecks : « vingt au sergent de",
    "ville, trois à Nastassia pour la lettre, j'en ai donc",
    "donné hier à Marmeladov quarante-sept ou",
    "cinquante », se dit-il. Il devait avoir une raison de",
    "calculer ainsi, mais il l'oublia en tirant l'argent de",
    "sa poche et ne s'en souvint qu'un peu plus tard en",
    "passant devant un marchand de comestibles, une",
    "sorte de gargote plutôt ; il sentit alors qu'il avait",
    "faim. "
]


src_3_fr = ["Ceci est un test de traduction avec un point. On se demande à quel point tout cela va bien fonctionner haha"]


from nllw.core import TranslationBackend
from matplotlib.ticker import MaxNLocator
from nllw.test_strings import src_2_fr
import matplotlib.pyplot as plt
import pandas as pd

src_texts = src_2_fr

colors = {
    "dark_gray_1": "#2E2E2E",
    "dark_gray_2": "#4B4B4B",
    "orange": "#FF7F0E",
    "red": "#D62728",
}


# print("1: direct generation)")
# translation_backend_1 = TranslationBackend(source_lang='fra_Latn', target_lang="eng_Latn")
# l_vals_no_cache = []

# for i in range(1, len(src_texts) + 1):
#     print(f'{i}/{len(src_texts) + 1}')
#     truncated_text = " ".join(src_texts[:i])
#     input_tokens = translation_backend_1.tokenizer(truncated_text, return_tensors="pt").to(translation_backend_1.device)
#     encoder_outputs = translation_backend_1.model.get_encoder()(**input_tokens)
#     output_tokens = translation_backend_1.model.generate(
#         encoder_outputs=encoder_outputs,
#         forced_bos_token_id=translation_backend_1.bos_token_id
#     )
#     output_text = translation_backend_1.tokenizer.decode(output_tokens[0], skip_special_tokens=True)
#     l_vals_no_cache.append({
#         "input": truncated_text,
#         "output_text": output_text,
#         "input_word_count": len(truncated_text.split()),
#         "output_word_count": len(output_text.split())
#     })

print("2: prefix reuse")
translation_backend_2 = TranslationBackend(source_lang='fra_Latn', target_lang="eng_Latn")
l_vals_with_cache = []

for i in range(1, len(src_texts) + 1):
    print(f'{i}/{len(src_texts) + 1}')
    truncated_text = " ".join(src_texts[:i])
    stable_translation, buffer = translation_backend_2.translate(truncated_text)
    
    full_output = stable_translation + buffer
    l_vals_with_cache.append({
        "input": truncated_text,
        "stable_translation": stable_translation,
        "buffer": buffer,
        "full_output": full_output,
        "input_word_count": len(truncated_text.split()),
        "stable_word_count": len(stable_translation.split()) if stable_translation else 0,
        "buffer_word_count": len(buffer.split()) if buffer else 0,
        "total_output_word_count": len(full_output.split()) if full_output else 0
    })

# df_base = pd.DataFrame(l_vals_no_cache)
df_prefix = pd.DataFrame(l_vals_with_cache)
# df_base.to_csv('output_base_analysis.csv')
df_prefix.to_csv('output_prefix_analysis.csv')

df_base = pd.read_csv('output_base_analysis.csv')
# df_prefix = pd.read_csv('output_prefix_analysis.csv')


iterations = list(range(1, len(df_prefix) + 1))
input_counts = df_prefix['input_word_count'].tolist()
output_with_prefix_and_buffer = df_prefix['total_output_word_count'].tolist()
stable_counts = df_prefix['stable_word_count'].tolist()
buffer_counts = df_prefix['buffer_word_count'].tolist()
base_output_wc = df_base['output_word_count'].tolist()

fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(14, 10))
fig, (ax1) = plt.subplots(1, 1, figsize=(5, 5))


ax1.plot(iterations, input_counts, label='Input', marker='o', color=colors['dark_gray_1'])
ax1.plot(iterations, base_output_wc, label='Base output (no prefix, no stability)', marker='^', color=colors['dark_gray_2'])
ax1.plot(iterations, stable_counts, label='With prefix (Stable part)', marker='^', linewidth=2, color=colors['red'])
ax1.plot(iterations, output_with_prefix_and_buffer, label='With prefix (Stable part + Buffer)', marker='^', color=colors['orange'])
ax1.set_xlabel('Iteration')
ax1.set_ylabel('Word count')
ax1.set_title('SimulMT french to english Word Count')
ax1.legend(fontsize=10)
ax1.grid(True, alpha=0.3)
ax1.xaxis.set_major_locator(MaxNLocator(integer=True))

# # Second plot: Detailed view of cached approach
# ax2.plot(iterations, input_counts, label='Input', marker='o', linewidth=2)
# ax2.plot(iterations, stable_counts, label='Stable prefix (cached)', marker='D', linewidth=2)
# ax2.plot(iterations, buffer_counts, label='Buffer (new)', marker='x', linewidth=2)
# ax2.plot(iterations, output_with_cache, label='Total output', marker='^', linewidth=2, linestyle='--')
# ax2.set_xlabel('Iteration', fontsize=12)
# ax2.set_ylabel('Word count', fontsize=12)
# ax2.set_title('Detailed view: Stable prefix vs Buffer (with cache)', fontsize=14, fontweight='bold')
# ax2.legend(fontsize=10)
# ax2.grid(True, alpha=0.3)

plt.tight_layout()
plt.savefig('translation_comparison.png', dpi=300, bbox_inches='tight')
plt.show()


import torch
import time
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from transformers.cache_utils import EncoderDecoderCache, DynamicCache
from typing import Tuple, Optional
from nllw.timed_text import TimedText

"""
nllb-200-distilled: pytorch_model.bin : 2.46 GB
"""

PUNCTUATION_MARKS = {'.', '!', '?', '。', '！', '？'}


class TranslationBackend:
    def __init__(
        self,
        source_lang,
        target_lang,
        model_name: str = "facebook/nllb-200-distilled-600M",
        model=None,
        tokenizer=None,
        verbose: bool = False,
        max_kept_sentences: int = 1
    ):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.verbose = verbose
        self.max_kept_sentences = max_kept_sentences

        if model is not None:
            self.model = model
            if not hasattr(model, 'device') or str(model.device) != self.device:
                self.model = self.model.to(self.device)
        else:
            self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name).to(self.device)

        if tokenizer is not None:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name, src_lang=source_lang)
        
        self.source_lang = source_lang
        self.target_lang = target_lang
        self.bos_token_id = self.tokenizer.convert_tokens_to_ids(self.target_lang)

        self.sentence_end_token_ids = set()

        # find all tokens that decode to sentence-ending punctuation. For ex: token 81 and 248075 both represent '.')
        for token_id in range(min(300000, self.tokenizer.vocab_size)):
            try:
                decoded = self.tokenizer.decode([token_id])
                cleaned = decoded.strip().strip("'\"").strip()
                if cleaned in PUNCTUATION_MARKS:
                    self.sentence_end_token_ids.add(token_id)
            except:
                pass

        self.input_buffer = []
        self.previous_tokens = []
        self.stable_prefix_segments = []
        self.stable_prefix_tokens = torch.tensor([], dtype=torch.int64)
        self.n_remaining_input_punctuation = 0

    def _trim(self) -> bool:
        x = 5
        # print(f'Before trimming: {len(self.input_buffer)} {len(self.stable_prefix_segments)}')
        self.input_buffer = self.input_buffer[-x-1:]
        self.stable_prefix_segments = self.stable_prefix_segments[-x:]
        # print(f'After trimming: {len(self.input_buffer)} {len(self.stable_prefix_segments)}')
    
    
    def simple_translation(self, text):
        inputs = self.tokenizer(text, return_tensors="pt").to(self.device)
        
        if self.verbose:
            start_time = time.time()
        encoder_outputs = self.model.get_encoder()(**inputs)
        if self.verbose:
            encoding_time = time.time() - start_time
        
        if self.verbose:
            start_time = time.time()
        output = self.generate(
                    encoder_outputs=encoder_outputs,
        )
        if self.verbose:
            decoding_time = time.time() - start_time
            print(f"Encoding time: {encoding_time:.4f}s | Decoding time: {decoding_time:.4f}s")
        
        result = self.tokenizer.decode(output[0], skip_special_tokens=True)
        return output, result

    def generate(
        self,
        encoder_outputs: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        max_length: Optional[int] = 200
        
    ) -> Tuple[torch.Tensor, Optional[Tuple]]:    
        with torch.no_grad():
            generated_tokens = self.model.generate(
                encoder_outputs=encoder_outputs,
                attention_mask=attention_mask,
                forced_bos_token_id=self.bos_token_id,
                max_length=max_length,
            )
        return generated_tokens

    def has_sentence_end_token(self, tokens):
        last_sentence_end = -1
        for i in range(len(tokens)):
            if tokens[i].item() in self.sentence_end_token_ids:
                last_sentence_end = i
        return last_sentence_end

    def compute_common_prefix_tokens(self):
        common_length = 0
        for i in range(min(len(self.previous_tokens), len(self.new_produced_tokens))):
            if self.previous_tokens[i] != self.new_produced_tokens[i]:
                common_length = i
                break
        else:
            common_length = min(len(self.previous_tokens), len(self.new_produced_tokens))

        last_sentence_end = self.has_sentence_end_token(self.new_produced_tokens[:common_length])

        if last_sentence_end >= 0:
            if self.n_remaining_input_punctuation > 0:
                print(f"\033[33mEOS detected. Remaining punctuation: {self.n_remaining_input_punctuation}\033[0m")
                self.n_remaining_input_punctuation -= 1
            else:
                print("\033[31mPrefix cut\033[0m")
                return self.new_produced_tokens[:last_sentence_end], self.new_produced_tokens[last_sentence_end:]
        return self.new_produced_tokens[:common_length], self.new_produced_tokens[common_length:]

    
    def translate(self, text: Optional[str | TimedText] = None) -> str:
        if type(text) == str:
            self.input_buffer.append(TimedText(text))
        elif type(text) == TimedText:
            self.input_buffer.append(text)
        
        # self._trim()
        buffer_text = ''.join([token.text for token in self.input_buffer])

        early_cut = False
        last_punct_pos = -1

        for i, char in enumerate(buffer_text):
            if char in PUNCTUATION_MARKS:
                last_punct_pos = i

        if last_punct_pos >= 0:
            text_to_process = buffer_text[:last_punct_pos + 1]
            remaining_text = buffer_text[last_punct_pos + 1:]

            char_count = 0
            split_index = len(self.input_buffer)
            for idx, timed_text in enumerate(self.input_buffer):
                char_count += len(timed_text.text)
                if char_count > last_punct_pos:
                    split_index = idx
                    if char_count - len(timed_text.text) <= last_punct_pos:
                        chars_before = last_punct_pos + 1 - (char_count - len(timed_text.text))
                        before_text = timed_text.text[:chars_before]
                        after_text = timed_text.text[chars_before:]
                        if after_text:
                            self.input_buffer = [TimedText(text=after_text, start=timed_text.start, end=timed_text.end)] + self.input_buffer[idx + 1:]
                        else:
                            self.input_buffer = self.input_buffer[idx + 1:]
                    break

            buffer_text = text_to_process
            early_cut = True
            print(f'\033[33mEarly cut. Processing: "{text_to_process}" | Remaining: "{remaining_text}"\033[0m')

        word_count = len(buffer_text.strip().split())
        if word_count < 3:
            return "", ""
        inputs = self.tokenizer(buffer_text, return_tensors="pt").to(self.device)
        self.n_remaining_input_punctuation += (self.has_sentence_end_token(inputs['input_ids'][0]) != -1)

        if self.stable_prefix_segments:
            self.stable_prefix_tokens = torch.cat(self.stable_prefix_segments, dim=0)
        if self.verbose:
            start_time = time.time()
        with torch.no_grad():
            encoder_outputs = self.model.get_encoder()(**inputs)
        if self.verbose:
            encoding_time = time.time() - start_time
        
        if self.verbose:
            start_time = time.time()
        if len(self.stable_prefix_tokens):
            # from bertviz import model_view
            # html_obj = model_view(output.attentions, source_tokens, html_action='return')
            # with open("attention_viz.html", "w") as f:
            #     f.write(str(html_obj.data))
            translation_tokens = self._continue_generation_with_cache(
                encoder_hidden_states=encoder_outputs.last_hidden_state,
            )
        else:
            with torch.no_grad():
                translation_tokens = self.generate(
                    encoder_outputs=encoder_outputs,
                    attention_mask=inputs['attention_mask'],
                )
        if self.verbose:
            decoding_time = time.time() - start_time
            print(f"Encoding time: {encoding_time:.4f}s | Decoding time: {decoding_time:.4f}s")

        self.new_produced_tokens = translation_tokens[0][len(self.stable_prefix_tokens):]
        if self.verbose:
            prefix_used_tokens = translation_tokens[0][:len(self.stable_prefix_tokens)]
            prefix_used = self.tokenizer.decode(
                prefix_used_tokens,
                skip_special_tokens=True
            )
        if early_cut:
            print('\033[33mResetting stable prefix state after early cut\033[0m')
            new_stable_tokens = self.new_produced_tokens
            self.stable_prefix_segments = []
            self.stable_prefix_tokens = torch.tensor([], dtype=torch.int64)
            self.previous_tokens = []
            self.n_remaining_input_punctuation = 0
            buffer = ''
        else:
            new_stable_tokens, new_buffer = self.compute_common_prefix_tokens()
            self.stable_prefix_segments.append(new_stable_tokens)
            self.previous_tokens = new_buffer
            buffer = self.tokenizer.decode(
                    new_buffer,
                    skip_special_tokens=True
            )
        stable_translation = self.tokenizer.decode(
            new_stable_tokens,
            skip_special_tokens=True
        )
        if self.verbose:
            # print(f'{i}/{len(src_texts) + 1}: \033[36m{truncated_text}\033[0m')
            print(f' \033[36m{prefix_used}\033[0m\033[32m{stable_translation}\033[0m \033[35m{buffer}\033[0m')
        return stable_translation, buffer

    def _continue_generation_with_cache(
        self,
        encoder_hidden_states: torch.Tensor,
        max_new_tokens: int = 200
    ) -> torch.Tensor:
        eos_token_id = self.tokenizer.eos_token_id

        with torch.no_grad():
            past_key_values = EncoderDecoderCache(
                DynamicCache(),
                DynamicCache()
            )

            decoder_out = self.model.model.decoder(
                input_ids=self.stable_prefix_tokens.unsqueeze(0),
                encoder_hidden_states=encoder_hidden_states,
                past_key_values=past_key_values,
                use_cache=True,
                return_dict=True,
            )
            past_key_values = decoder_out.past_key_values
            prefix_logits = self.model.lm_head(decoder_out.last_hidden_state)
            next_token_id = torch.argmax(prefix_logits[:, -1, :], dim=-1).unsqueeze(-1)

            generated_tokens = self.stable_prefix_tokens.unsqueeze(0).clone()

            for _ in range(max_new_tokens):
                if next_token_id.item() == eos_token_id:
                    break

                generated_tokens = torch.cat([generated_tokens, next_token_id], dim=-1)

                decoder_out = self.model.model.decoder(
                    input_ids=next_token_id,
                    encoder_hidden_states=encoder_hidden_states,
                    past_key_values=past_key_values,
                    use_cache=True,
                    return_dict=True,
                )
                past_key_values = decoder_out.past_key_values
                logits = self.model.lm_head(decoder_out.last_hidden_state)
                next_token_id = torch.argmax(logits[:, -1, :], dim=-1).unsqueeze(-1)

        return generated_tokens



if __name__ == '__main__':
    from nllw.test_strings import *
    import pandas as pd
    translation_backend = TranslationBackend(source_lang='fra_Latn', target_lang="eng_Latn", verbose=True)
    # input_text = " ".join(src_2_fr)
    src_texts = src_2_fr 

    l_vals_with_cache = []
    for i in range(0, len(src_texts)):
        truncated_text = src_texts[i]
        print(f'\n\n{i}/{len(src_texts) + 1}: {truncated_text}')
        stable_translation, buffer = translation_backend.translate(truncated_text)
        # print(f'\033[32m{stable_translation}\033[0m \033[35m{buffer}\033[0m')
        
        full_output = stable_translation + buffer
        l_vals_with_cache.append({
            "input": truncated_text,
            "stable_translation": stable_translation,
            # "buffer_tokens": translation_backend.buffer_tokens,
            "stable_prefix_tokens": translation_backend.stable_prefix_tokens,
            "input_word_count": len(truncated_text.split()),
            "stable_word_count": len(stable_translation.split()) if stable_translation else 0,
            "total_output_word_count": len(full_output.split()) if full_output else 0
        })
    pd.DataFrame(l_vals_with_cache).to_pickle('export_with_tokens.pkl')